from une_ai.models import Agent, GridMap

class TicTacToePlayer(Agent):

    def __init__(self, agent_name, agent_program, board_size=3):
        self._board_size = board_size
        super().__init__(agent_name, agent_program)

    def is_valid_box(self, v):
        if v is None:
            return True
        
        return isinstance(v, tuple) and len(v) == 2 and v[0] in range(0,self._board_size) and v[1] in range(0,self._board_size)

    def add_all_sensors(self):
        self.add_sensor(
            'game-board-sensor',
            GridMap(self._board_size, self._board_size, None),
            lambda v, s=self._board_size: isinstance(v, GridMap) and v.get_height() == s and v.get_width() == s
        )

        self.add_sensor(
            'turn-taking-indicator',
            'X',
            lambda v: v in ['X', 'O']
        )
    
    def add_all_actuators(self):
        self.add_actuator(
            'marker',
            None,
            self.is_valid_box
        )
    
    def add_all_actions(self):
        for i in range(0, self._board_size):
            for j in range(0, self._board_size):
                self.add_action(
                    'mark-{0}-{1}'.format(i, j),
                    lambda r=i, c=j: {'marker': (r, c)} if self.is_valid_box((r, c)) else {}
                )


import random
import numpy as np
from scipy.signal import convolve2d, convolve

from une_ai.models import GameEnvironment
from une_ai.models import GridMap

class InvalidMoveException(Exception):
    pass

class ConnectFourBaseEnvironment(GameEnvironment):

    POWERUPS = ['anvil', 'wall', 'x2']
    N_COLS = 7
    N_ROWS = 6

    def __init__(self):
        super().__init__("Connect Four Game Environment")
        self._game_board = GridMap(ConnectFourBaseEnvironment.N_COLS, ConnectFourBaseEnvironment.N_ROWS, None)
        self._played_powerup = (None, None)
        self._player_turn = 'Y' # yellow always start
    
    def add_player(self, player):
        assert len(self._players) < 2, "It is not possible to add more than 2 players for this game."

        if len(self._players) == 0:
            colour = 'Y'
        else:
            colour = 'R'

        powerup = random.choice(ConnectFourBaseEnvironment.POWERUPS)
        
        self._players[colour] = (player, powerup) # adding player
        return colour
    
    def get_game_state(self):
        game_state = {
            'game-board': self._game_board.copy(),
            'power-up-Y': None if 'Y' not in self._players.keys() else self._players['Y'][1],
            'power-up-R': None if 'R' not in self._players.keys() else self._players['R'][1],
            'player-turn': self._player_turn
        }

        return game_state
    
    def turn(game_state):
        return game_state['player-turn']
    
    def get_openings(game_board, player_colour):
        opponent_colour = 'R' if player_colour == 'Y' else 'Y'
        
        # 7 comes from 2*(n_to_connect - 1) + 1
        kernel = np.array([ 1, 1, 1, 1])
        max_diagonals = game_board.get_width() + game_board.get_height() - 7
        diagonal_idx = lambda normalised_idx: normalised_idx - int(max_diagonals / 2) + 1
        
        game_board_map = game_board.get_map()
        rows = {
            'horizontals': [r for r in game_board_map],
            'verticals': [game_board_map[:, i] for i in range(game_board.get_width())],
            'left-diagonals': [np.diagonal(game_board_map, diagonal_idx(i)) for i in range(max_diagonals)],
            'right-diagonals': [np.diagonal(np.fliplr(game_board_map), diagonal_idx(i)) for i in range(max_diagonals)]
        }

        def fill(row):
            flipped = np.flip(row)
            mode = False
            filled = flipped.copy()
            for i, item in enumerate(flipped):
                if item == player_colour:
                    mode = True
                elif item == opponent_colour:
                    mode = False
                
                if mode == True and item is None:
                    filled[i] = player_colour
            
            return np.flip(filled)

        results = {}
        for k, v in rows.items():
            results[k] = []
            for i, row in enumerate(v):
                row_optimal = fill(row)
                if k == 'horizontals':
                    row_optimal = fill(np.flip(row_optimal))
                
                values = convolve(row == player_colour, kernel, mode="valid")
                values_optimal = convolve(row_optimal == player_colour, kernel, mode="valid")
                
                is_free = True if 4 in values_optimal else False
                if is_free:
                    max_val = np.max(values)
                    if max_val > 0:
                        if k == 'horizontals':
                            idx = ConnectFourBaseEnvironment.N_ROWS - 1 - i
                            results[k].append((idx , max_val))
                        else:
                            results[k].append((i, max_val))

        return results
    
    def get_winner(game_state):
        game_board = game_state['game-board']

        horizontal_kernel = np.array([[ 1, 1, 1, 1]])
        vertical_kernel = np.transpose(horizontal_kernel)
        diag_kernel = np.eye(4, dtype=np.uint8)
        flipped_diag_kernel = np.fliplr(diag_kernel)
        detection_kernels = [horizontal_kernel, vertical_kernel, diag_kernel, flipped_diag_kernel]

        winners = []
        for player in ['Y', 'R']:
            player_checkers = game_board.get_map() == player
            for kernel in detection_kernels:
                convolved_values = convolve2d(player_checkers, kernel, mode="valid")
                if (convolved_values == 4).any():
                    if player not in winners:
                        winners.append(player)
        
        if len(winners) == 2:
            # both players got 4 connected because of a popup move
            # wins the player of the current turn
            return game_state['player-turn']
        elif len(winners) == 1:
            return winners[0]
        else:
            return None
    
    def get_percepts(self):
        game_state = self.get_game_state()
        return {
            'game-board-sensor': game_state['game-board'],
            'powerups-sensor': {
                'Y': game_state['power-up-Y'],
                'R': game_state['power-up-R'],
            },
            'turn-taking-indicator': self._player_turn
        }
    
    def get_first_free_row(game_board, column_num):
        column = game_board.get_column(column_num)
        for i in range(len(column)-1, -1, -1):
            if column[i] is None:
                return i
        
        return None
    
    def is_valid_column(game_board, col_num):
        return col_num is not None and col_num >= 0 and col_num < game_board.get_width()
    
    def is_column_full(game_board, col_num):
        return ConnectFourBaseEnvironment.get_first_free_row(game_board, col_num) == None
    
    def _change_player_turn(self):
        self._prev_turn = self._player_turn
        if self._player_turn == 'Y':
            self._player_turn = 'R'
        else:
            self._player_turn = 'Y'
    
    def _use_powerup(self, player_colour):
        powerup = self._players[player_colour][1]
        self._players[player_colour][1] = None
        self._played_powerup = (player_colour, powerup)
    
    def transition_result(game_state, action):
        assert action.startswith('use-power-up-') or action.startswith('release-') or action.startswith('popup-'), "Invalid action '{0}'.".format(action)
        try:
            index = int(action.split('-')[-1])
        except:
            index = None
        game_board = game_state['game-board'].copy()
        assert index is not None and index >= 0 and index < game_board.get_width(), "The column index {0} from the action {1} is not a valid index.".format(index, action)

        def release_token(game_board, column_num, checker):
            row = ConnectFourBaseEnvironment.get_first_free_row(game_board, column_num)
            game_board.set_item_value(column_num, row, checker)

        new_game_state = {
            'game-board': game_board,
            'power-up-Y': game_state['power-up-Y'],
            'power-up-R': game_state['power-up-R'],
            'player-turn': game_state['player-turn']
        }

        colour_turn = game_state['player-turn']
        if action.startswith('use-power-up-'):
            # playing turn with a power up
            powerup = game_state['power-up-{0}'.format(colour_turn)]
            if powerup is None:
                raise InvalidMoveException("The player of colour '{0}' does not have a power-up available.".format(colour_turn))
            if powerup not in ConnectFourBaseEnvironment.POWERUPS:
                raise InvalidMoveException("The power-up '{0}' is not a valid power-up value.".format(powerup))
            if not ConnectFourBaseEnvironment.is_valid_column(game_board, index):
                raise InvalidMoveException("Invalid column index {0}. When playing the {1} power up, the agent must use the 'checker-releaser' actuator to select a valid column where to play the power up.".format(index, powerup))
            if ConnectFourBaseEnvironment.is_column_full(game_board, index) and powerup != 'anvil':
                raise InvalidMoveException("The column at index {0} is full".format(index))
            
            # remove powerup because used
            new_game_state['power-up-{0}'.format(colour_turn)] = None

            if powerup == 'anvil':
                for i in range(0, game_board.get_height()):
                    game_board.set_item_value(index, i, None)
                release_token(game_board, index, colour_turn)
                new_game_state['player-turn'] = 'Y' if colour_turn == 'R' else 'R'
                return new_game_state
            elif powerup == 'wall':
                release_token(game_board, index, 'W')
                new_game_state['player-turn'] = 'Y' if colour_turn == 'R' else 'R'
                return new_game_state
            elif powerup == 'x2':
                release_token(game_board, index, colour_turn)
                return new_game_state
        elif action.startswith('popup-'):
            # player decided to popup a checker
            if not ConnectFourBaseEnvironment.is_valid_column(game_board, index):
                raise InvalidMoveException("Invalid column index {0}. When popping out a checker, the actuator 'checker-grabber' must select a valid column index from where to popup the checker.".format(index))
            bottom_idx = game_board.get_height()-1
            bottom_checker = game_board.get_item_value(index, bottom_idx)
            if bottom_checker is None:
                raise InvalidMoveException("The column at index {0} is empty. It is not possible to pop-out checkers from there.".format(index))
            if bottom_checker != colour_turn:
                raise InvalidMoveException("A player can pop-out only their checkers from the bottom of the column. The bottom checker at column index {0} is {1}, whereas the player colour is {2}.".format(index, bottom_checker, colour_turn))
            
            for i in range(bottom_idx, 0, -1):
                top_checker = game_board.get_item_value(index, i-1)
                game_board.set_item_value(index, i, top_checker)
            game_board.set_item_value(index, 0, None)
            new_game_state['player-turn'] = 'Y' if colour_turn == 'R' else 'R'
            return new_game_state
        elif action.startswith('release-'):
            # player decided to add a checker
            if not ConnectFourBaseEnvironment.is_valid_column(game_board, index):
                raise InvalidMoveException("Invalid column index {0}. When releasing a checker, the actuator 'checker-releaser' must select a valid column index where to release the checker.".format(index))
            if ConnectFourBaseEnvironment.is_column_full(game_board, index):
                raise InvalidMoveException("The column at index {0} is full".format(index))
            
            release_token(game_board, index, colour_turn)
            new_game_state['player-turn'] = 'Y' if colour_turn == 'R' else 'R'
            return new_game_state
        
        return new_game_state
    
    def state_transition(self, agent_actuators):
        assert 'checker-handler' not in agent_actuators.keys() or (isinstance(agent_actuators['checker-handler'], tuple) and len(agent_actuators['checker-handler']) == 2), "The actuator 'checker-handler' takes a tuple (handling_type, column_index) as value."

        if 'checker-handler' not in agent_actuators.keys():
            # no transition possible
            return
        
        powerup = agent_actuators['powerup-selector']
        handling_type, column_num = agent_actuators['checker-handler']
        if handling_type not in ['release', 'popup']:
            raise InvalidMoveException("The first value of the tuple in the actuator 'checker-handler' must be either 'release' or 'popup'.")
        
        if powerup == True and handling_type != 'release':
            raise InvalidMoveException("When playing a power up with the 'powerup-selector' actuator set to True, the agent must also use the 'checker-handler' actuator to select a valid column where to release the power up. Instead, the handling type for the 'checker-handler' was set to '{0}'.".format(handling_type))

        action = None
        if powerup == True:
            action = 'use-power-up-{0}'.format(column_num)
        elif handling_type == 'release':
            action = 'release-{0}'.format(column_num)
        elif handling_type == 'popup':
            action = 'popup-{0}'.format(column_num)

        new_state = ConnectFourBaseEnvironment.transition_result(self.get_game_state(), action)

        self._game_board = new_state['game-board'].copy()
        self._players['Y'] = (self._players['Y'][0], new_state['power-up-Y'])
        self._players['R'] = (self._players['Y'][0], new_state['power-up-R'])
        self._player_turn = new_state['player-turn']


import pygame

from une_ai.models import GridMap

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
RED = (255, 0, 0)
BLUE = (0, 0, 255)

DISPLAY_WIDTH = 320
DISPLAY_HEIGHT = 240
SCORE_BAR_HEIGHT = 40
BOX_SIZE = 50

class TicTacToeGame:

    def __init__(self, agent_X, agent_O, environment, board_size=3, display_w=DISPLAY_WIDTH, display_h=DISPLAY_HEIGHT, box_size=BOX_SIZE):
        assert type(agent_X).__name__ == 'TicTacToePlayer', "agent_X must be an instance of a the class TicTacToePlayer"
        assert type(agent_O).__name__ == 'TicTacToePlayer', "agent_O must be an instance of the class TicTacToePlayer"
        assert type(environment).__name__ == 'TicTacToeGameEnvironment', "The parameter environment must be an instance of the class TicTacToeGameEnvironment"
        
        pygame.init()
        window = pygame.display.set_mode((DISPLAY_WIDTH, DISPLAY_HEIGHT+SCORE_BAR_HEIGHT))
        pygame.display.set_caption('Tic Tac Toe')
        self._window_clock = pygame.time.Clock()

        self._box_size = box_size
        self._display = window
        self._display_size = (display_w, display_h)
        self._agents = {}
        self._agents['X'] = agent_X
        self._agents['O'] = agent_O
        self._environment = environment
        self._board_size = board_size
        
        fonts = pygame.font.get_fonts()
        self._font = fonts[0] # default to a random font
        # try to look among the most common fonts
        test_fonts = ['arial', 'couriernew', 'verdana', 'helvetica', 'roboto']
        for font in test_fonts:
            if font in fonts:
                self._font = font
                break

        self.main()
        
    def _play_step(self):
        game_state = self._environment.get_game_state()
        if type(self._environment).is_terminal(game_state):
            return
        
        try:
            cur_marker = type(self._environment).turn(game_state)
        except:
            print("Error. The game environment class does not return a correct game state. Skipping updating the game state")
            return
        
        # SENSE
        self._agents[cur_marker].sense(self._environment)
        # THINK
        actions = self._agents[cur_marker].think()
        # ACT
        self._agents[cur_marker].act(actions, self._environment)
    
    def _reset_bg(self):
        self._display.fill(BLACK)
    
    def _draw_box(self, x, y, color, alpha = 255):
        font = pygame.font.SysFont(self._font, 30)
        game_state = self._environment.get_game_state()
        n_cells = self._board_size
        padding_left = int((self._display_size[0] - n_cells*self._box_size)/2)
        padding_top = int((self._display_size[1] - n_cells*self._box_size)/2)
        x_coord = padding_left + x*self._box_size
        y_coord = padding_top + y*self._box_size

        surface = pygame.Surface((self._box_size-2,self._box_size-2))
        surface.set_alpha(alpha) # alpha level
        pygame.draw.rect(surface, color, surface.get_rect())
        self._display.blit(surface, (x_coord, y_coord) )

        if isinstance(game_state, dict) and 'game-board' in game_state.keys():
            game_board = game_state['game-board']
        else:
            print("Error. The game environment class does not return a correct game state. Drawing an empty game board.")
            game_board = GridMap(self._board_size, self._board_size, None)

        cur_mark = game_board.get_item_value(x, y)
        if cur_mark != None:
            color = RED if cur_mark == 'X' else BLUE
            text_size = font.size(cur_mark)
            mark_text = font.render(cur_mark, True, color)
            top = y_coord + int((BOX_SIZE - text_size[1])/2)
            left = x_coord + int((BOX_SIZE - text_size[0])/2)
            self._display.blit(mark_text, (left, top))
    
    def _draw_board(self):
        n_cells = self._board_size
        for i in range(0, n_cells):
            for j in range(0, n_cells):
                self._draw_box(j, i, WHITE)
    
    def _draw_game_over(self):
        winner = type(self._environment).get_winner(self._environment.get_game_state())
        
        if winner is not None:
            text = "Player {0} won!".format(winner)
        else:
            text = "Tie!"
        
        n_cells = self._board_size
        font = pygame.font.SysFont(self._font, 20)
        text_size = font.size(text)
        game_over_text = font.render(text, True, WHITE)
        padding_top = int((self._display_size[1] - n_cells*self._box_size)/2)
        x_coord = int((self._display_size[0] - text_size[0])/2)
        y_coord = padding_top + n_cells*self._box_size + 30
        self._display.blit(game_over_text, (x_coord, y_coord))
                
    def _draw_frame(self):
        self._reset_bg()
        self._draw_board()
        if type(self._environment).is_terminal(self._environment.get_game_state()):
            self._draw_game_over()

    def main(self):
        running = True

        while running:
            # update frame
            self._draw_frame()
            # Update the clock and display
            pygame.display.update()

            #Event Tasking
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    running = False
                    quit()
            
            # updating game with one step
            # sense - think - act
            self._play_step()
            
            self._window_clock.tick(1)



from .graph_node import GraphNode

class MCTSGraphNode(GraphNode):

    def __init__(self, state, parent_node, action):
        super().__init__(state, parent_node, action, 1)
        self._number_of_visits = 0
        self._wins = {}
        
        self._successors = {}
    
    def get_successors(self):
        return list(self._successors.values())
    
    def was_action_expanded(self, action):
        return action in self._successors.keys()
    
    def wins(self, player):
        if player in self._wins.keys():
            return self._wins[player]
        else:
            return 0
    
    def n(self):
        return self._number_of_visits
    
    def add_successor(self, successor_state, action):
        if not self.was_action_expanded(action):
            successor = MCTSGraphNode(successor_state, self, action)
            self._successors[action] = successor

        return self._successors[action]
    
    def is_leaf_node(self):
        return self.n() == 0
    
    def backpropagate(self, winner):
        self._number_of_visits += 1.
        if winner in self._wins.keys():
            self._wins[winner] += 1.
        elif winner is not None:
            self._wins[winner] = 0
        parent = self.get_parent_node()
        if parent is not None:
            parent.backpropagate(winner)


import numpy as np

class GridMap:
    def __init__(self, w, h, default_value=False):
        self._w = w
        self._h = h
        self._default_value = default_value
        self._map = np.array([[default_value]*w]*h) #np.zeros((h, w), dtype=dtype)

    def get_width(self):
        return self._w
    
    def get_height(self):
        return self._h
    
    def set_item_value(self, x, y, value):
        assert x >= 0 and x < self._w, "Coordinate x={0} out of bounds (w={1})".format(x, self._w)
        assert y >= 0 and y < self._h, "Coordinate y={0} out of bounds (h={1})".format(y, self._h)

        self._map[y, x] = value
    
    def get_item_value(self, x, y):
        assert x >= 0 and x < self._w, "Coordinate x={0} out of bounds (w={1})".format(x, self._w)
        assert y >= 0 and y < self._h, "Coordinate y={0} out of bounds (h={1})".format(y, self._h)

        return self._map[y, x]
    
    def get_row(self, row_idx):
        assert row_idx >= 0 and row_idx < self._h, "Row index = {0} out of bounds (h={1})".format(row_idx, self._h)

        return self._map[row_idx, :]
    
    def get_column(self, col_idx):
        assert col_idx >= 0 and col_idx < self._w, "Column index = {0} out of bounds (w={1})".format(col_idx, self._w)

        return self._map[:, col_idx]
    
    def get_map(self):
        return self._map.copy()
    
    def find_value_by_condition(self, condition_function):
        rc_coords = np.argwhere(condition_function(self._map)).tolist()
        xy_coords = []
        for coord in rc_coords:
            xy_coords.append((coord[1], coord[0]))
        
        return xy_coords

    def find_value(self, value):
        return self.find_value_by_condition(lambda gridmap: gridmap == value)
    
    def set_map(self, new_map):
        assert new_map.shape[0] == self._h and new_map.shape[1] == self._w, "The new map does not have the same dimensions of the original map. Original map dimensions {0}, new map dimensions {1}.".format(self._map.shape, new_map.shape)
        self._map = new_map.copy()

    def copy(self):
        new_grid = GridMap(self._w, self._h, self._default_value)
        new_grid.set_map(self._map)

        return new_grid


from une_ai.assignments import SnakeEnvironment

import pygame
import math
import random

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)
YELLOW = (255, 255, 0)

pygame.init()
DISPLAY_WIDTH = 640
DISPLAY_HEIGHT = 480
SCORE_BAR_HEIGHT = 40
GAME_OVER_HEIGHT = 40
TILE_SIZE = 10

window = pygame.display.set_mode((DISPLAY_WIDTH, DISPLAY_HEIGHT+SCORE_BAR_HEIGHT))
pygame.display.set_caption('Assignment 1 - Snake')
window_clock = pygame.time.Clock()

class SnakeGame:

    def __init__(self, snake_agent, display_w=DISPLAY_WIDTH, display_h=DISPLAY_HEIGHT, tile_size=TILE_SIZE, max_time=60, tick=5):
        self._tile_size = tile_size
        self._display = window
        self._display_size = (display_w, display_h)
        w_env = math.floor(display_w / self._tile_size)
        h_env = math.floor(display_h / self._tile_size)
        self._environment = SnakeEnvironment(w_env, h_env, max_time)
        self._agent = snake_agent
        # update sensors with initial state
        self._agent.sense(self._environment)
        self._tick = tick

        fonts = pygame.font.get_fonts()
        self._font = fonts[0] # default to a random font
        # try to look among the most common fonts
        test_fonts = ['arial', 'couriernew', 'verdana', 'helvetica', 'roboto']
        for font in test_fonts:
            if font in fonts:
                self._font = font
                break

        self.main()
        
    def _play_step(self):
        # SENSE
        self._agent.sense(self._environment)
        # THINK
        actions = self._agent.think()
        # ACT
        self._agent.act(actions, self._environment)
    
    def _reset_bg(self):
        self._display.fill(BLACK)
    
    def _draw_tile(self, x, y, color, alpha = 255):
        x_coord = x*self._tile_size
        y_coord = y*self._tile_size

        surface = pygame.Surface((self._tile_size,self._tile_size))
        surface.set_alpha(alpha) # alpha level
        pygame.draw.rect(surface, color, surface.get_rect())
        self._display.blit(surface, (x_coord, y_coord,self._tile_size,self._tile_size) )
    
    def _draw_snake(self):
        if self._environment.is_game_over() and self._environment.get_time() > 0:
            color = YELLOW
        else:
            color = GREEN
        percepts = self._environment.get_percepts()
        snake_body = percepts['body-sensor']
        for segment in snake_body:
            self._draw_tile(segment[0], segment[1], color)

    def _draw_food(self):
        food_coords = self._environment.get_food_coords()
        for coord in food_coords:
            self._draw_tile(coord[0], coord[1], RED)
    
    def _draw_walls(self):
        walls_coords = self._environment.get_walls_coords()
        for coord in walls_coords:
            self._draw_tile(coord[0], coord[1], WHITE)
        
        # draw as a wall the rest of the window outside the boundings
        # of the environment (if not multiple of the tile size)
        if self._display_size[0] % self._tile_size > 0:
            x_coord = self._environment.get_width()*(self._tile_size - 1)
            y_coord = 0
            pygame.draw.rect(self._display, WHITE, pygame.Rect(x_coord, y_coord, self._display_size[0] - x_coord, self._display_size[1]))
        if self._display_size[1] % self._tile_size > 0:
            x_coord = 0
            y_coord = self._environment.get_height()*(self._tile_size - 1)
            pygame.draw.rect(self._display, WHITE, pygame.Rect(x_coord, y_coord, self._display_size[0], self._display_size[1] - y_coord))
    
    def _draw_score_bar(self):
        surface = pygame.Surface((DISPLAY_WIDTH,SCORE_BAR_HEIGHT))
        pygame.draw.rect(surface, WHITE, surface.get_rect())
        self._display.blit(surface, (0, DISPLAY_HEIGHT,DISPLAY_WIDTH,SCORE_BAR_HEIGHT) )
        font = pygame.font.SysFont(self._font, 15)
        score_text_str = "Score: {0}".format(self._environment.get_score())
        score_text_size = font.size(score_text_str)
        score_text = font.render(score_text_str, True, BLACK)
        if self._environment.is_game_over():
            time = 0
        else:
            time = int(self._environment.get_time())
        time_text_str = "Time: {0}".format(time)
        time_text_size = font.size(time_text_str)
        time_text = font.render(time_text_str, True, BLACK)
        padding_top = int((SCORE_BAR_HEIGHT - score_text_size[1]) / 2)
        white_space = DISPLAY_WIDTH - score_text_size[0] - time_text_size[0]
        padding_side = int(white_space * 0.25)
        padding_central = white_space - 2 * padding_side
        self._display.blit(score_text, (padding_side, DISPLAY_HEIGHT+padding_top))
        self._display.blit(time_text, (padding_side + score_text_size[0] + padding_central, DISPLAY_HEIGHT+padding_top))
    
    def _draw_game_over(self):
        surface = pygame.Surface((DISPLAY_WIDTH,DISPLAY_HEIGHT))
        surface.set_alpha(60)
        pygame.draw.rect(surface, WHITE, surface.get_rect())
        y_pos = int((DISPLAY_HEIGHT-GAME_OVER_HEIGHT)/2)
        self._display.blit(surface, (0, 0, DISPLAY_WIDTH,DISPLAY_HEIGHT) )
        font = pygame.font.SysFont(self._font, 30)
        game_over_text = font.render("GAME OVER", True, RED)
        self._display.blit(game_over_text, (int((DISPLAY_WIDTH-150)/2), int((DISPLAY_HEIGHT-30)/2), 100, 30))

    def _draw_frame(self):
        self._reset_bg()
        self._draw_walls()
        self._draw_food()
        self._draw_snake()
        self._draw_score_bar()
        if self._environment.is_game_over():
            self._draw_game_over()

    def main(self):
        running = True

        while running:
            # update frame
            self._draw_frame()

            #Event Tasking
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    running = False
                    quit()
            
            # updating game with one step
            # sense - think - act
            if not self._environment.is_game_over():
                self._play_step()

            # Update the clock and display
            pygame.display.update()
            window_clock.tick(self._tick)
            elapsed = window_clock.get_time() / 1000
            self._environment.tick(elapsed)



import uuid
import os
import json
from abc import ABC, abstractmethod
import logging


from une_ai.models import GraphNode

class TTable(ABC):

    ZOBRIST_TABLE_FILENAME = 'zobrist-table.json'
    TT_FILENAME = 'transposition-table.json'

    def __init__(self, saving_directory='./transposition-tables/', instance_id=None, verbose=False):
        if instance_id is None:
            self._instance_id = str(uuid.uuid1())
        else:
            self._instance_id = instance_id
        
        self._zobrist_table = None
        self._tt = None
        self._tt_keys = []

        self._saving_dir = saving_directory
        if not os.path.exists(self.get_saving_directory()):
            os.makedirs(self.get_saving_directory())
        
        self._zobrist_table_filename = os.path.join(self._saving_dir, self._instance_id, TTable.ZOBRIST_TABLE_FILENAME)
        self._tt_filename = os.path.join(self._saving_dir, self._instance_id, TTable.TT_FILENAME)

        self._verbose = verbose
        self._logger = None

        if self._verbose:
            logging.basicConfig()
            logging.root.setLevel(logging.NOTSET)
            self._logger = logging.getLogger("TTable({0})".format(self.get_instance_id()))

        self._init_zobrist_table()
        self._init_transposition_table()
    
    @abstractmethod
    def generate_zobrist_table(self):
        pass

    @abstractmethod
    def compute_hash(self, state):
        pass

    def set_verbose(self, is_verbose):
        self._verbose = is_verbose
    
    def size(self):
        return len(self._tt['transposition-table'].keys())
    
    def get_instance_id(self):
        return self._instance_id
    
    def get_zobrist_table_filename(self):
        return self._zobrist_table_filename
    
    def get_ttable_filename(self):
        return self._tt_filename
    
    def get_saving_directory(self):
        return os.path.join(self._saving_dir, self._instance_id)
    
    def log(self, message):
        if self._verbose:
            self._logger.info(message)
    
    def _init_zobrist_table(self):
        if os.path.exists(self._zobrist_table_filename):
            self.log("Found existing zobrist table, loading it")
            with open(self._zobrist_table_filename, 'r') as f:
                self._zobrist_table = json.load(f)
        else:
            self.log("Zobrist table not found, creating one")
            self._zobrist_table = self.generate_zobrist_table()
            with open(self._zobrist_table_filename, 'w') as f:
                json.dump(self._zobrist_table, f)
    
    def _init_transposition_table(self):
        if os.path.exists(self._tt_filename):
            self.log("Found existing transposition table, loading it")
            with open(self._tt_filename, 'r') as f:
                self._tt = json.load(f)
            self._tt_keys = list(self._tt['transposition-table'].keys())
        else:
            self.log("Transposition table not found, creating one")
            self._tt = {
                'transposition-table': {},
                'instance-id': self._instance_id,
                'saving-dir': self._saving_dir,
                'zobrist-table-file': self._zobrist_table_filename,
                'tt-file': self._tt_filename
            }
            self._save_to_file()
    
    def get_zobrist_table(self):
        if self._zobrist_table is None:
            self._init_zobrist_table()
        
        return self._zobrist_table.copy()
    
    def get_ttable(self):
        if self._tt is None:
            self._init_transposition_table()
        
        return self._tt.copy()
    
    def store_node(self, node, entry_dict):
        assert isinstance(node, GraphNode), "The parameter node must be an instance of GraphNode"
        assert isinstance(entry_dict, dict), "The parameter entry_dict must be a dictionary"
        assert self._tt is not None, "The transposition table was not initialised yet"

        state_hash = self.compute_hash(node.get_state())

        # check if the node state is already stored in the tt table
        any_changes = state_hash not in self._tt_keys
        if state_hash in self._tt_keys:
            cur_entry = self._tt['transposition-table'][state_hash]
            cur_entry_dict = cur_entry['dict-values']
            origin_keys = list(cur_entry_dict.keys()).sort()
            updated_keys = list(entry_dict.keys()).sort()
            if origin_keys == updated_keys:
                for k, value in cur_entry_dict.items():
                    if value != entry_dict[k]:
                        any_changes = True
                        break
        
        if any_changes:
            self.log("Storing new node with state hash '{0}', values = {1}".format(state_hash, entry_dict))
            if state_hash not in self._tt['transposition-table'].keys():
                self._tt['transposition-table'][state_hash] = {}
                self._tt_keys.append(state_hash)
            self._tt['transposition-table'][state_hash]['dict-values'] = entry_dict
            self._save_to_file()

    def lookup(self, node):
        assert isinstance(node, GraphNode), "The parameter node must be an instance of GraphNode"

        state_hash = self.compute_hash(node.get_state())

        if state_hash in self._tt_keys:
            # node state already stored, retrieve it
            tt = self.get_ttable()
            return tt['transposition-table'][state_hash]['dict-values']
        
        return None
    
    def _save_to_file(self):
        assert self._tt is not None, "The transposition table was not initialised yet"

        with open(self._tt_filename, 'w') as f:
            json.dump(self._tt, f)
            self.log("New version of the table successfully saved at '{0}' (size = {1})".format(self._tt_filename, self.size()))
    


from .agent import Agent
from .environment import Environment
from .grid_map import GridMap
from .graph_node import GraphNode
from .game_environment import GameEnvironment
from .transposition_table import TTable
from .MCTS_graph_node import MCTSGraphNode


import numpy as np
import random
import math
import logging

from une_ai.models import Environment, GridMap

class InvalidMoveException(Exception):
    pass

class SnakeEnvironment(Environment):

    def __init__(self, width, height, max_time):
        super().__init__("Snake Environment")

        self._w = width
        self._h = height
        self._food_map = GridMap(width, height, 0)
        self._walls_map = GridMap(width, height)
        self._initialise_walls(0.01)
        self._initialise_food()
        empty_coords = self.get_free_coords()
        agent_pos = random.choice(empty_coords)
        self._snake = [agent_pos]
        self._grow = False
        self._prev_direction = None

        self._score = 0
        self._clock = max_time
        self._game_over = False
    
    def _initialise_walls(self, wall_density=0.01):
        total_size = self._w*self._h
        n_walls = math.floor(total_size*wall_density)
        empty_tiles = self._walls_map.find_value(False)
        walls_coords = random.sample(empty_tiles, n_walls)
        for coord in walls_coords:
            orientation = random.choice(['h', 'v'])
            length = random.choice([3, 4, 5])
            cur_x = coord[0]
            cur_y = coord[1]
            for i in range(0, length):
                if cur_x >= self._w or cur_y >= self._h:
                    #if out of bounds, end
                    break
                self._walls_map.set_item_value(cur_x, cur_y, True)
                if orientation == 'h':
                    cur_x += 1
                else:
                    cur_y += 1
    
    def _initialise_food(self):
        food = [5, 10, 20]
        self._spawn_food(food)
            
    def _spawn_food(self, food):
        condition_func = lambda walls_map: np.logical_and(walls_map == False, self._food_map.get_map() == 0)
        valid_coords = self._walls_map.find_value_by_condition(condition_func)
        new_coord = random.sample(valid_coords, len(food))
        for i, coord in enumerate(new_coord):
            self._food_map.set_item_value(coord[0], coord[1], food[i])
    
    def get_food_coords(self):
        condition_func = lambda food_map: food_map != 0
        return self._food_map.find_value_by_condition(condition_func)
    
    def get_walls_coords(self):
        return self._walls_map.find_value(True)
    
    def get_free_coords(self):
        condition_func = lambda walls_map: np.logical_and(walls_map == False, self._food_map.get_map() == 0)
        return self._walls_map.find_value_by_condition(condition_func)
    
    def move_snake_head(self, direction):
        assert direction in ['up', 'down', 'left', 'right'], "Direction not valid ('{0}').".format(direction)

        if self.is_game_over():
            return
        
        if (self._prev_direction == 'up' and direction == 'down') or (self._prev_direction == 'down' and direction == 'up') or (self._prev_direction == 'left' and direction == 'right') or (self._prev_direction == 'right' and direction == 'left'):
            raise(InvalidMoveException("Invalid move action! The current direction of the head snake is '{0}' and it cannot go the opposite direction '{1}'. You might be missing something when adding your actions...".format(self._prev_direction, direction)))
        
        self._prev_direction = direction

        prev_body = self._snake.copy()

        # move the head according to the current direction
        cur_head = self._snake[0]
        if direction == 'up':
            new_head = (cur_head[0], cur_head[1] - 1)
        elif direction == 'down':
            new_head = (cur_head[0], cur_head[1] + 1)
        elif direction == 'left':
            new_head = (cur_head[0] - 1, cur_head[1])
        elif direction == 'right':
            new_head = (cur_head[0] + 1, cur_head[1])
        
        self._snake.insert(0, new_head)

        # remove the snake's tail if it did not eat fruit
        if not self._grow:
            self._snake.pop()
        else:
            # stop growing
            self._grow = False

        if self.did_snake_collide():
            self._snake = prev_body
            self._game_over = True
    
    def did_snake_collide(self):
        head = self._snake[0]
        try:
            is_obstacle = self._walls_map.get_item_value(head[0], head[1]) or head in self._snake[1:]
        except:
            # out of bounds, crashed against border
            is_obstacle = True
        
        return is_obstacle
    
    def eat_food(self):
        head = self._snake[0]
        food = self._food_map.get_item_value(head[0], head[1])
        if food != 0:
            self._food_map.set_item_value(head[0], head[1], 0)
            self._spawn_food([food])
            self._score += food
            self._grow = True
        else:
            # eating where there is no food, penalty
            self._score -= 5

    def is_game_over(self):
        return self._game_over
    
    def get_score(self):
        return self._score
    
    def tick(self, tick):
        self._clock -= tick
        if self._clock <= 0:
            self._clock = 0
            self._game_over = True
    
    def get_time(self):
        return self._clock
    
    def get_percepts(self):
        percepts = {}

        percepts['body-sensor'] = self._snake.copy()

        food_coords = self.get_food_coords()
        food = []
        for coord in food_coords:
            points = self._food_map.get_item_value(coord[0], coord[1])
            food.append((coord[0], coord[1], points))
        percepts['food-sensor'] = food

        percepts['obstacles-sensor'] = self._walls_map.find_value(True)
        percepts['clock'] = int(self._clock)

        return percepts
    
    def state_transition(self, agent_actuators):
        if agent_actuators is not None and 'head' in agent_actuators.keys():
            try:
                direction = agent_actuators['head']
                if direction in ['up', 'down', 'left', 'right']:
                    self.move_snake_head(direction)
                else:
                    logging.error("The value '{0}' is not a valid direction for the 'head' actuator.".format(direction))
            except AssertionError as err:
                logging.warning("The actuator 'head' might not be implemented, it might be implemented incorrectly or an invalid move was chosen. Error: {0}".format(err))
        else:
            logging.warning("The actuator 'head' might not be implemented or implemented incorrectly.")
        
        if agent_actuators is not None and 'mouth' in agent_actuators.keys():
            try:
                mouth_state = agent_actuators['mouth']
                if mouth_state in ['open', 'close']:
                    if mouth_state == 'open':
                        self.eat_food()
                else:
                    logging.error("The value '{0}' is not a valid mouth state for the 'mouth' actuator.".format(mouth_state))
            except Exception as err:
                logging.warning("The actuator 'mouth' might not be implemented or it might be implemented incorrectly. Error: {0}".format(err))
        else:
            logging.warning("The actuator 'mouth' might not be implemented or implemented incorrectly.")


import pygame
import numpy as np

BLACK = (30, 30, 30)
WHITE = (255, 255, 255)
RED = (255, 0, 0)
YELLOW = (255, 255, 0)
GRAY = (125, 125, 125)
BLUE = (10, 20, 200)
GREEN = (255, 125, 125)

DISPLAY_WIDTH = 640
DISPLAY_HEIGHT = 500
POWERUPS_BAR_HEIGHT = 35
SCORE_BAR_HEIGHT = 35
INFO_BAR_HEIGHT = 35
BOX_SIZE = 60

class ConnectFourGame:

    def __init__(self, agent_Y, agent_R, environment, is_debugging=False, display_w=DISPLAY_WIDTH, display_h=DISPLAY_HEIGHT, box_size=BOX_SIZE):
        assert type(agent_R).__name__ == 'ConnectFourPlayer', "agent_R must be an instance of the class ConnectFourPlayer"
        assert type(agent_Y).__name__ == 'ConnectFourPlayer', "agent_Y must be an instance of the class ConnectFourPlayer"
        assert type(environment).__name__ == 'ConnectFourEnvironment', "environment must be an instance of a subclass of the class ConnectFourBaseEnvironment"

        pygame.init()
        window = pygame.display.set_mode((DISPLAY_WIDTH, DISPLAY_HEIGHT+SCORE_BAR_HEIGHT))
        pygame.display.set_caption('Connect Four')
        window_clock = pygame.time.Clock()

        self._box_size = box_size
        self._display = window
        self._window_clock = window_clock
        self._display_size = (display_w, display_h)
        self._agents = {}
        self._agents['Y'] = agent_Y
        self._agents['R'] = agent_R
        self._environment = environment
        self._last_action = ""
        self._is_debugging = is_debugging

        game_state = self._environment.get_game_state()
        game_board = game_state['game-board']
        self._n_cols = game_board.get_width()
        self._n_rows = game_board.get_height()
        self._padding_left = int((self._display_size[0] - self._n_cols*self._box_size)/2)
        self._padding_top = int((self._display_size[1] - self._n_rows*self._box_size)/2)

        self._coordinates = np.array([[None]*self._n_cols]*self._n_rows)
        for y in range(self._n_rows):
            for x in range(self._n_cols):
                self._coordinates[y, x] = (x, y)

        fonts = pygame.font.get_fonts()
        self._font = fonts[0] # default to a random font
        # try to look among the most common fonts
        test_fonts = ['arial', 'couriernew', 'verdana', 'helvetica', 'roboto']
        for font in test_fonts:
            if font in fonts:
                self._font = font
                break

        self.main()
        
    def _play_step(self):
        game_state = self._environment.get_game_state()
        if type(self._environment).is_terminal(game_state):
            return
        
        cur_colour = type(self._environment).turn(game_state)
        
        # SENSE
        self._agents[cur_colour].sense(self._environment)
        # THINK
        actions = self._agents[cur_colour].think()
        player = 'Yellow' if cur_colour == 'Y' else 'Red'
        if len(actions) != 0:
            self._last_action = "{0} player played the move '{1}'".format(player, actions[0])
        # ACT
        self._agents[cur_colour].act(actions, self._environment)
    
    def _reset_bg(self):
        self._display.fill(BLACK)
    
    def _draw_box(self, x, y, color, alpha = 255):
        game_state = self._environment.get_game_state()
        game_board = game_state['game-board']
        x_coord = self._padding_left + x*self._box_size
        y_coord = self._padding_top + y*self._box_size

        surface = pygame.Surface((self._box_size,self._box_size))
        surface.set_alpha(alpha) # alpha level
        pygame.draw.rect(surface, color, surface.get_rect())
        self._display.blit(surface, (x_coord, y_coord))

        cur_checker = game_board.get_item_value(x, y)
        
        if cur_checker == 'Y':
            checker_color = YELLOW
        elif cur_checker == 'R':
            checker_color = RED
        elif cur_checker == 'W':
            checker_color = GRAY
        else:
            # no checker
            checker_color = WHITE
        checker_surface = pygame.Surface((self._box_size,self._box_size))
        checker_surface.set_alpha(255) # alpha level
        checker_surface.fill(color)
        radius = int((self._box_size/2)*0.8)
        center = int(self._box_size/2)
        pygame.draw.circle(checker_surface, checker_color, (center, center), radius)
        self._display.blit(checker_surface, (x_coord, y_coord))
    
    def _draw_board(self):
        game_state = self._environment.get_game_state()
        game_board = game_state['game-board']
        winner = type(self._environment).get_winner(game_state)
        winning_coords = []
        max_diagonals = game_board.get_width() + game_board.get_height() - 7
        diagonal_idx = lambda normalised_idx: normalised_idx - int(max_diagonals / 2) + 1
        if winner is not None:
            openings = type(self._environment).get_openings(game_board, winner)
            for k, l_openings in openings.items():
                for t in l_openings:
                    if t[1] == 4:
                        if k == 'horizontals':
                            winning_coords = self._coordinates[game_board.get_height() -1 - t[0], :].tolist()
                            break
                        elif k == 'verticals':
                            winning_coords = self._coordinates[:, t[0]].tolist()
                            break
                        elif k == 'left-diagonals':
                            winning_coords = np.diagonal(self._coordinates, diagonal_idx(t[0])).tolist()
                            break
                        else:
                            winning_coords = np.diagonal(np.fliplr(self._coordinates), diagonal_idx(t[0])).tolist()
                            break
        for i in range(0, self._n_cols):
            for j in range(0, self._n_rows):
                if (i, j) in winning_coords:
                    color = GREEN
                else:
                    color = BLUE
                self._draw_box(i, j, color)
    
    def _draw_text(self, text_message, padding_top, orientation, font_size = 20):
        font = pygame.font.SysFont(self._font, font_size)
        text_size = font.size(text_message)
        text = font.render(text_message, True, WHITE)
        top = self._padding_top + self._n_rows*self._box_size + 10 + padding_top
        if orientation == 'center':
            left = int((self._display_size[0] - text_size[0])/2)
        elif orientation == 'left':
            left = self._padding_left
        elif orientation == 'right':
            left = self._display_size[0] - text_size[0] - self._padding_left
        else:
            left = 0
        self._display.blit(text, (left, top))
    
    def _draw_game_over(self):
        game_state = self._environment.get_game_state()
        game_board = game_state['game-board']
        colour = type(self._environment).get_winner(game_state)
        if colour == 'Y':
            winner = 'Yellow player'
        elif colour == 'R':
            winner = 'Red player'
        else:
            winner = None
        
        if winner is not None:
            text = "{0} won!".format(winner)
        else:
            text = "Tie!"
        
        padding_top = POWERUPS_BAR_HEIGHT + INFO_BAR_HEIGHT
        self._draw_text(text, padding_top, 'center')
    
    def _draw_powerups(self):
        game_state = self._environment.get_game_state()
        powerup_y = game_state['power-up-Y']
        powerup_r = game_state['power-up-R']

        self._draw_text("Yellow player power-up: {0}".format(powerup_y), 0, 'left', 15)
        self._draw_text("Red player power-up: {0}".format(powerup_r), 0, 'right', 15)
    
    def _draw_frame(self):
        self._reset_bg()
        self._draw_board()
        self._draw_powerups()
        self._draw_text("Last action: {0}".format(self._last_action), POWERUPS_BAR_HEIGHT, 'left', 15)
        if type(self._environment).is_terminal(self._environment.get_game_state()):
            self._draw_game_over()
        else:
            colour = type(self._environment).turn(self._environment.get_game_state())
            player = "Yellow" if colour == 'Y' else "Red"
            self._draw_text("Player Turn: {0}".format(player), POWERUPS_BAR_HEIGHT + INFO_BAR_HEIGHT, 'left', 15)
    
    def key_to_action(key):
        if key >= pygame.K_1 and key <= pygame.K_7:
            col = key - pygame.K_1
            return 'release-{0}'.format(col)
        else:
            for col, k in enumerate(['q', 'w', 'e', 'r', 't', 'y', 'u']):
                if key == getattr(pygame, 'K_{0}'.format(k)):
                    return 'popup-{0}'.format(col)
            for col, k in enumerate(['a', 's', 'd', 'f', 'g', 'h', 'j']):
                if key == getattr(pygame, 'K_{0}'.format(k)):
                    return 'use-power-up-{0}'.format(col)
        
        return None
    
    def wait_for_user_input():
        action = None
        while action is None:
            event = pygame.event.wait()
            if event.type == pygame.KEYDOWN:
                action = ConnectFourGame.key_to_action(event.key)
        
        return action

    def main(self):
        running = True

        while running:
            # update frame
            self._draw_frame()
            pygame.display.update()
            self._window_clock.tick(1)
            #Event Tasking
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    running = False
                    quit()

            
            # updating game with one step
            # sense - think - act
            self._play_step()

            if self._is_debugging:
                pygame.time.delay(2000)
            



from une_ai.models import Agent

import pygame
import math

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)
YELLOW = (255, 255, 0)
MAGENTA = (255, 0, 255)

pygame.init()
DISPLAY_WIDTH = 320
DISPLAY_HEIGHT = 240
SCORE_BAR_HEIGHT = 40
TILE_SIZE = 10

window = pygame.display.set_mode((DISPLAY_WIDTH, DISPLAY_HEIGHT+SCORE_BAR_HEIGHT))
pygame.display.set_caption('Vacuum cleaner agent')
window_clock = pygame.time.Clock()

class VacuumGame:

    def __init__(self, vacuum_agent, uses_charging_dock=False, display_w=DISPLAY_WIDTH, display_h=DISPLAY_HEIGHT, tile_size=TILE_SIZE):
        assert isinstance(vacuum_agent, Agent), "The parameter vacuum_agent must be an instance of a subclass of the class Agent."

        self._tile_size = tile_size
        self._display = window
        self._display_size = (display_w, display_h)
        w_env = math.floor(display_w / self._tile_size)
        h_env = math.floor(display_h / self._tile_size)
        if uses_charging_dock:
            from une_ai.vacuum import VacuumDockEnvironment
            self._environment = VacuumDockEnvironment(w_env, h_env)
        else:
            from une_ai.vacuum import VacuumEnvironment
            self._environment = VacuumEnvironment(w_env, h_env)
        self._agent = vacuum_agent

        # updating the sensors based on initial environment state
        self._agent.sense(self._environment)

        fonts = pygame.font.get_fonts()
        self._font = fonts[0] # default to a random font
        # try to look among the most common fonts
        test_fonts = ['arial', 'couriernew', 'verdana', 'helvetica', 'roboto']
        for font in test_fonts:
            if font in fonts:
                self._font = font
                break

        self.main()
        
    def _play_step(self):        
        # SENSE
        self._agent.sense(self._environment)
        # THINK
        actions = self._agent.think()
        # ACT
        self._agent.act(actions, self._environment)
    
    def _reset_bg(self):
        self._display.fill(BLACK)
    
    def _draw_tile(self, x, y, color, alpha = 255):
        x_coord = x*self._tile_size
        y_coord = y*self._tile_size

        surface = pygame.Surface((self._tile_size,self._tile_size))
        surface.set_alpha(alpha) # alpha level
        pygame.draw.rect(surface, color, surface.get_rect())
        self._display.blit(surface, (x_coord, y_coord,self._tile_size,self._tile_size) )
    
    def _draw_charging_dock(self):
        try:
            x, y = self._environment.get_charging_dock_location()
            self._draw_tile(x, y, MAGENTA)
        except:
            # using the environment without charging dock
            pass
    
    def _draw_vacuum_agent(self):
        color = GREEN
        p = self._environment.get_percepts()
        pos = (p['location-sensor'][0], p['location-sensor'][1])
        agent_state = self._environment.get_agent_state()
        if agent_state['collided'] is not None:
            color = YELLOW

        self._draw_tile(pos[0], pos[1], color)

    def _draw_dirt(self):
        dirt_coords = self._environment.get_dirt_coords()
        for coord in dirt_coords:
            self._draw_tile(coord[0], coord[1], RED)
    
    def _draw_walls(self):
        walls_coords = self._environment.get_walls_coords()
        for coord in walls_coords:
            self._draw_tile(coord[0], coord[1], WHITE)
        
        # draw as a wall the rest of the window outside the boundings
        # of the environment (if not multiple of the tile size)
        if self._display_size[0] % self._tile_size > 0:
            x_coord = self._environment.get_width()*(self._tile_size - 1)
            y_coord = 0
            pygame.draw.rect(self._display, WHITE, pygame.Rect(x_coord, y_coord, self._display_size[0] - x_coord, self._display_size[1]))
        if self._display_size[1] % self._tile_size > 0:
            x_coord = 0
            y_coord = self._environment.get_height()*(self._tile_size - 1)
            pygame.draw.rect(self._display, WHITE, pygame.Rect(x_coord, y_coord, self._display_size[0], self._display_size[1] - y_coord))
    
    def _draw_explored(self):
        explored = self._environment.get_explored_tiles_coords()
        for coord in explored:
            self._draw_tile(coord[0], coord[1], BLUE, 80)
    
    def _draw_score_bar(self):
        surface = pygame.Surface((DISPLAY_WIDTH,SCORE_BAR_HEIGHT))
        pygame.draw.rect(surface, WHITE, surface.get_rect())
        self._display.blit(surface, (0, DISPLAY_HEIGHT,DISPLAY_WIDTH,SCORE_BAR_HEIGHT) )
        font = pygame.font.SysFont(self._font, 15)
        score_text_str = "Score: {0}".format(self._environment.get_score())
        score_text_size = font.size(score_text_str)
        score_text = font.render(score_text_str, True, BLACK)
        battery_text_str = "Battery level: {0}%".format(self._agent.get_battery_level())
        battery_text_size = font.size(battery_text_str)
        battery_text = font.render(battery_text_str, True, BLACK)
        white_space = DISPLAY_WIDTH - score_text_size[0] - battery_text_size[0]
        padding_top = int((SCORE_BAR_HEIGHT - score_text_size[1]) / 2)
        padding_side = int(white_space * 0.25)
        central_padding = white_space - (padding_side * 2)
        self._display.blit(score_text, (padding_side, DISPLAY_HEIGHT+padding_top))
        self._display.blit(battery_text, (padding_side + score_text_size[0] + central_padding, DISPLAY_HEIGHT+padding_top))

    def _draw_frame(self):
        self._reset_bg()
        self._draw_walls()
        self._draw_explored()
        self._draw_dirt()
        self._draw_charging_dock()
        self._draw_vacuum_agent()
        self._draw_score_bar()

    def main(self):
        running = True

        while running:
            # update frame
            self._draw_frame()

            #Event Tasking
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    running = False
                    quit()
            
            # updating game with one step
            # sense - think - act
            self._play_step()

            # Update the clock and display
            pygame.display.update()
            window_clock.tick(5)



import numpy as np
import random
import math
import logging

from une_ai.models.grid_map import GridMap
from une_ai.models.environment import Environment

class VacuumEnvironment(Environment):

    def __init__(self, w, h):
        super().__init__('Vacuum world')
        self._w = w
        self._h = h
        self._dirt_map = GridMap(w, h)
        self._walls_map = GridMap(w, h)
        self._explored_tiles = GridMap(w, h)
        self._initialise_walls(0.01)
        self._initialise_dirt(0.01)
        empty_coords = self.get_free_coords()
        agent_pos = random.choice(empty_coords)
        self._max_battery_level = 1000
        self._agent_state = {
            'position': agent_pos,
            'battery': self._max_battery_level,
            'collided': None
        }

        self._score = 0
    
    def get_percepts(self):
        # starts the percepts with current values of sensors
        percepts = {}
        agent_state = self.get_agent_state()
        agent_x, agent_y = agent_state['position']
        coords = {
            'center': (agent_x, agent_y),
            'north': (agent_x, agent_y-1),
            'south': (agent_x, agent_y+1),
            'west': (agent_x-1, agent_y),
            'east': (agent_x+1, agent_y)
        }

        # reset bumper sensors
        for direction in coords.keys():
            if direction != 'center':
                percepts['bumper-sensor-{0}'.format(direction)] = False
        
        # check if the agent crashed against a wall
        collision_dir = agent_state['collided']
        if collision_dir is not None:
            percepts['bumper-sensor-{0}'.format(collision_dir)] = True
        
        percepts['location-sensor'] = agent_state['position']
        percepts['battery-level'] = (agent_state['battery'] / self._max_battery_level)*100

        # now check if there is dirt in the surroundings
        for coord_name, coord in coords.items():
            try:
                is_dirty = self.is_dirty(coord[0], coord[1])
            except:
                is_dirty = False
            percepts['dirt-sensor-{0}'.format(coord_name)] = is_dirty
            
        return percepts
    
    def get_agent_state(self):
        return self._agent_state.copy()
    
    def _consume_agent_battery(self, consumption):
        agent_state = self.get_agent_state()
        cur_level = agent_state['battery']
        if cur_level - consumption < 0:
            new_level = 0
        else:
            new_level = cur_level - consumption
        
        self._agent_state['battery'] = new_level

    def state_transition(self, agent_actuators):
        agent_state = self.get_agent_state()
        try:
            direction = agent_actuators['wheels-direction']
        except:
            # wheels-direction actuator not implemented, using north as default direction
            logging.warning("The actuator 'wheels-direction' is not implemented. Using default direction 'north'")
            direction = 'north'
        
        try:
            suction_power = agent_actuators['suction-power']
        except:
            # suction-power actuator not implemented, using default value
            logging.warning("The actuator 'suction-power' is not implemented. Using default value 0")
            suction_power = 0
        
        try:
            vacuum_power = agent_actuators['vacuum-power']
        except:
            # vacuum-power actuator not implemented, using default value
            logging.warning("The actuator 'vacuum-power' is not implemented. Using default value 0")
            vacuum_power = 0
        
        cur_x, cur_y = agent_state['position']
        try:
            self.set_explored(cur_x, cur_y)
        except:
            # out of boundaries, but that's all good
            # the environment class will manage the
            # crash with the wall in the following instructions
            pass
        
        # remove dirt on current position if suction power on
        if agent_state['battery'] > 0 and suction_power == 1:
            self.remove_dirt(cur_x, cur_y)
            self._consume_agent_battery(5)

        # moving the agent in the new position
        if agent_state['battery'] > 0 and vacuum_power == 1:
            self._consume_agent_battery(1)
            if direction == "west":
                offset = (-1, 0)
            elif direction == "east":
                offset = (1, 0)
            elif direction == "north":
                offset = (0, -1)
            elif direction == "south":
                offset = (0, 1)
            
            new_x, new_y = (cur_x + offset[0], cur_y + offset[1])

            # reset bumper sensors
            self._agent_state['collided'] = None
            
            # check that the agent is not crashing against a wall
            if not self.is_wall(new_x, new_y):
                # the position is a valid position
                self._agent_state['position'] = (new_x, new_y)
            else:
                # there is a wall!
                # update collision state
                self._agent_state['collided'] = direction
                # set the wall position as explored
                try:
                    self._explored_tiles.set_item_value(new_x, new_y, True)
                except:
                    # out of bounds, pass
                    pass
    
    def get_score(self):
        return self._score

    def get_width(self):
        return self._w
    
    def get_height(self):
        return self._h
    
    def set_explored(self, pos_x, pos_y):
        self._explored_tiles.set_item_value(pos_x, pos_y, True)
        
    def _initialise_walls(self, wall_density=0.01):
        total_size = self._w*self._h
        n_walls = math.floor(total_size*wall_density)
        empty_tiles = self._walls_map.find_value(False)
        walls_coords = random.sample(empty_tiles, n_walls)
        for coord in walls_coords:
            orientation = random.choice(['h', 'v'])
            length = random.choice([3, 4, 5])
            cur_x = coord[0]
            cur_y = coord[1]
            for i in range(0, length):
                if cur_x >= self._w or cur_y >= self._h:
                    #if out of bounds, end
                    break
                self._walls_map.set_item_value(cur_x, cur_y, True)
                if orientation == 'h':
                    cur_x += 1
                else:
                    cur_y += 1
    
    def _initialise_dirt(self, dirt_density=0.01):
        total_size = self._w*self._h
        n_dirt = math.floor(total_size*dirt_density)
        self._spawn_dirt(n_dirt)
            
    def _spawn_dirt(self, k=1):
        condition_func = lambda walls_map: np.logical_and(walls_map == False, self._dirt_map.get_map() == False)
        valid_coords = self._walls_map.find_value_by_condition(condition_func)
        new_coord = random.sample(valid_coords, k)
        for coord in new_coord:
            self._dirt_map.set_item_value(coord[0], coord[1], True)

    def is_dirty(self, pos_x, pos_y):
        return self._dirt_map.get_item_value(pos_x, pos_y)
    
    def is_wall(self, pos_x, pos_y):
        try:
            return self._walls_map.get_item_value(pos_x, pos_y)
        except:
            # out of bounds, it's a wall
            return True
    
    def remove_dirt(self, x, y):
        try:
            is_dirt = self.is_dirty(x, y)
            if is_dirt:
                self._score += 1
                self._dirt_map.set_item_value(x, y, False)
        except:
            # we may be out of bounds, but there is no dirt outside bounds
            pass
    
    def get_dirt_coords(self):
        return self._dirt_map.find_value(True)
    
    def get_walls_coords(self):
        return self._walls_map.find_value(True)
    
    def get_free_coords(self):
        condition_func = lambda walls_map: np.logical_and(walls_map == False, self._dirt_map.get_map() == False)
        return self._walls_map.find_value_by_condition(condition_func)
    
    def get_explored_tiles_coords(self):
        return self._explored_tiles.find_value(True)


from abc import ABC, abstractmethod, abstractstaticmethod
from .environment import Environment
from .agent import Agent

class GameEnvironment(Environment):

    def __init__(self, environment_name):
        super().__init__(environment_name)

        self._players = {}
    
    @abstractmethod
    def add_player(self, player):
        pass

    def get_player(self, player_name):
        assert player_name in self._players.keys(), "The player with name '{0}' was not added to the game.".format(player_name)

        return self._players[player_name]
    
    @abstractmethod
    def get_game_state(self):
        pass

    @abstractstaticmethod
    def get_legal_actions(game_state):
        pass

    @abstractstaticmethod
    def is_terminal(game_state):
        pass
    
    @abstractstaticmethod
    def get_winner(game_state):
        pass

    @abstractstaticmethod
    def transition_result(game_state, action):
        pass

    @abstractstaticmethod
    def turn(game_state):
        pass

    @abstractstaticmethod
    def payoff(game_state, player_name):
        pass


from une_ai.vacuum import VacuumEnvironment

class VacuumDockEnvironment(VacuumEnvironment):

    def __init__(self, w, h):
        super().__init__(w, h)

        self._dock_location = self._agent_state['position']
        self._max_battery_level = 300
    
    def get_charging_dock_location(self):
        return self._dock_location
    
    def get_percepts(self):
        percepts = super().get_percepts()
        percepts['charging-dock-location-sensor'] = self.get_charging_dock_location()

        return percepts
    
    def recharge(self):
        self._agent_state['battery'] = self._max_battery_level
    
    def state_transition(self, agent_actuators):
        agent_state = self.get_agent_state()
        agent_pos = agent_state['position']
        if agent_pos == self.get_charging_dock_location():
            self.recharge()
        return super().state_transition(agent_actuators)
    
    


import logging
from abc import ABC, abstractmethod

class Agent(ABC):

    def __init__(self, agent_name, agent_program):
        assert callable(agent_program), "The parameter 'agent_program' must be a function: f(Percepts) -> Actions"

        self._agent_name = agent_name
        self._sensors = {}
        self._actuators = {}
        self._actions = {}
        self._agent_program = agent_program

        self.add_all_sensors()
        self.add_all_actuators()
        self.add_all_actions()
    
    def who_am_I(self):
        return self._agent_name
    
    # ABSTRACT methods

    @abstractmethod
    def add_all_sensors(self):
        pass

    @abstractmethod
    def add_all_actuators(self):
        pass

    @abstractmethod
    def add_all_actions(self):
        pass

    # SENSORS methods
    
    def add_sensor(self, sensor_name, initial_value, validation_function):
        assert sensor_name not in self._sensors.keys(), "The sensor '{0}' already exists for this agent.".format(sensor_name)
        assert callable(validation_function), "The parameter validation_function must be callable."
        assert validation_function(initial_value), "The initial_value {0} set for the sensor '{1}' is not a valid value.".format(initial_value, sensor_name)

        self._sensors[sensor_name] = {
            'value': initial_value,
            'validation-function': validation_function
        }
    
    def _update_sensor_value(self, sensor_name, sensor_value):
        assert sensor_name in self._sensors.keys(), "The sensor '{0}' does not exist.".format(sensor_name)
        assert self._sensors[sensor_name]['validation-function'](sensor_value), "The value {0} is not a valid value for the sensor '{1}'".format(sensor_value, sensor_name)

        self._sensors[sensor_name]['value'] = sensor_value
    
    def sense(self, environment):
        percepts = environment.get_percepts()
        for sensor_name, percept in percepts.items():
            try:
                self._update_sensor_value(sensor_name, percept)
            except Exception as err:
                # sensor does not exist or value is invalid
                logging.warning(err)
    
    def read_sensor_value(self, sensor_name):
        assert sensor_name in self._sensors.keys(), "The sensor '{0}' is not a valid sensor's name".format(sensor_name)
        return self._sensors[sensor_name]['value']
    
    def read_sensors(self):
        sensors = {}
        for sensor_name in self._sensors.keys():
            sensors[sensor_name] = self.read_sensor_value(sensor_name)
        
        return sensors
    
    # Agent thinking behaviour methods

    def think(self):
        actions = self._agent_program(self.read_sensors(), self.read_actuators())
        return actions

    # ACTUATORS / ACTIONS methods

    def add_actuator(self, actuator_name, initial_value, validation_function):
        assert actuator_name not in self._actuators.keys(), "The actuator '{0}' already exists for this agent.".format(actuator_name)
        assert callable(validation_function), "The parameter validation_function must be callable."
        assert validation_function(initial_value), "The initial_value {0} set for the actuator '{1}' is not a valid value.".format(initial_value, actuator_name)
        
        self._actuators[actuator_name] = {
            'value': initial_value,
            'validation-function': validation_function
        }
    
    def _update_actuator_value(self, actuator_name, actuator_value):
        assert actuator_name in self._actuators.keys(), "The actuator '{0}' does not exist.".format(actuator_name)
        assert self._actuators[actuator_name]['validation-function'](actuator_value), "The value {0} is not a valid value for the actuator '{1}'".format(actuator_value, actuator_name)

        self._actuators[actuator_name]['value'] = actuator_value
    
    def read_actuator_value(self, actuator_name):
        assert actuator_name in self._actuators.keys(), "The actuator '{0}' is not a valid actuator's name".format(actuator_name)

        return self._actuators[actuator_name]['value']
    
    def read_actuators(self):
        actuators = {}
        for actuator_name in self._actuators.keys():
            actuators[actuator_name] = self.read_actuator_value(actuator_name)
        
        return actuators
    
    def add_action(self, action_name, action_function):
        assert action_name not in self._actions.keys(), "The action '{0}' already exists for this agent.".format(action_name)
        assert callable(action_function), "The parameter action_function must be callable."
        
        self._actions[action_name] = action_function

    def act(self, actions, environment):
        if actions is not None:
            # updates the actuators based on the 
            # chosen actions and given the action_function
            for action in actions:
                if action not in self._actions.keys():
                    raise KeyError("The action '{0}' is not a valid action for this agent.".format(action))
                else:
                    updated_actuators = self._actions[action]()
                    for actuator_name, actuator_value in updated_actuators.items():
                        self._update_actuator_value(actuator_name, actuator_value)
        
        environment.state_transition(self.read_actuators())

        return True

    


from abc import ABC, abstractmethod

class Environment(ABC):

    def __init__(self, environment_name):
        self._name = environment_name
    
    def get_environment_name(self):
        return self._name
    
    @abstractmethod
    def get_percepts(self):
        pass

    @abstractmethod
    def state_transition(self, agent_actuators):
        pass


class GraphNode():

    def __init__(self, state, parent_node, action, cost=0):
        assert parent_node is None or isinstance(parent_node, GraphNode), "The parent_node parameter must be a GraphNode instance or None."
        self._state = state
        self._parent_node = parent_node
        self._action = action
        self._cost = cost
        self._path = None
        self._path_cost = None
    
    def get_parent_node(self):
        return self._parent_node
    
    def get_cost(self):
        return self._cost
    
    def get_state(self):
        return self._state
    
    def get_action(self):
        return self._action
    
    def get_path(self):
        if self._path is not None and self._path_cost is not None:
            return self._path, self._path_cost
        
        parent_node = self.get_parent_node()
        path_cost = self.get_cost()
        action = self.get_action()
        if parent_node is not None:
            path = [self.get_action()]
        else:
            path = []
        if parent_node is not None:
            path_parent, path_cost_parent = parent_node.get_path()
            path_cost += path_cost_parent
            path = path_parent + path

        self._path = path
        self._path_cost = path_cost
        
        return path, path_cost

    
